import pytest
from httpx import Response
from api_client.async_client import OVCirrusApiClient
from models.organization import Organization
from models.generic import ApiResponse
from api_client.auth import Authenticator
from unittest.mock import AsyncMock
from typing import Optional

# Dummy Authenticator that always returns a token
class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"

    def force_relogin(self) -> bool:
        return True

# Mock response data for organization settings
mock_org_settings_data = {
    "status": 200,
    "message": "Success",
    "data": {
        "id": "123",
        "name": "ALE",
        "createdAt": "2022-08-04T12:10:38.058Z",
        "updatedAt": "2022-09-08T15:56:53.407Z",
        "enforceStrongPassword": True,
        "enforceStrongPasswordNotifyType": "SHOW_MESSAGE_AFTER_LOGIN",
        "timezone": "Europe/Tirane",
        "auditHour": 130,
        "idleTimeout": 3600
    }
}

# Mock invalid response (e.g., 400 Bad Request)
mock_invalid_response = {
    "status": 400,
    "message": "Bad Request",
    "data": None
}

# Mock no response (None)
mock_no_response = None

@pytest.mark.asyncio
async def test_get_organization_basic_settings():
    # Mock the 'get' method of OVCirrusApiClient to return the mock valid response
    mock_get = AsyncMock(return_value=mock_org_settings_data)

    # Create an instance of OVCirrusApiClient with the mock get method
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.get = mock_get  # Mocking the actual API call

    # Define the organization ID
    org_id = "123"

    # Run the method
    response: Optional[ApiResponse[Organization]] = await client.getOrganizationBasicSettings(org_id)

    # Assertions for valid response
    assert response is not None
    assert response.status == 200
    assert response.data.id == org_id
    assert response.data.name == "ALE"
    assert response.data.enforceStrongPassword is True
    assert response.data.timezone == "Europe/Tirane"
    assert response.data.auditHour == 130
    assert response.data.idleTimeout == 3600

    # Close the client after the test
    await client.close()


@pytest.mark.asyncio
async def test_get_organization_invalid_status():
    # Mock the 'get' method of OVCirrusApiClient to return the mock invalid response
    mock_get = AsyncMock(return_value=mock_invalid_response)

    # Create an instance of OVCirrusApiClient with the mock get method
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.get = mock_get  # Mocking the actual API call

    # Define the organization ID
    org_id = "123"

    # Run the method
    response: Optional[ApiResponse[Organization]] = await client.getOrganizationBasicSettings(org_id)

    # Assertions for invalid response
    assert response is not None
    assert response.status == 400
    assert response.message == "Bad Request"
    assert response.data is None

    # Close the client after the test
    await client.close()


@pytest.mark.asyncio
async def test_get_organization_no_response():
    # Mock the 'get' method of OVCirrusApiClient to return None (no response)
    mock_get = AsyncMock(return_value=mock_no_response)

    # Create an instance of OVCirrusApiClient with the mock get method
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.get = mock_get  # Mocking the actual API call

    # Define the organization ID
    org_id = "123"

    # Run the method
    response: Optional[ApiResponse[Organization]] = await client.getOrganizationBasicSettings(org_id)

    # Assertions for no response
    assert response is None

    # Close the client after the test
    await client.close()



from pydantic import BaseModel  # Use BaseModel directly
from typing import Optional, TypeVar, Generic  # Make sure to import the correct types

T = TypeVar("T")

class ApiResponse(BaseModel, Generic[T]):  # Inherit directly from BaseModel
    status: Optional[int] = None
    message: Optional[str] = None
    data: Optional[T] = None




import pytest
from unittest.mock import AsyncMock
from api_client.async_client import OVCirrusApiClient
from models.organization import Organization
from models.generic import ApiResponse
from datetime import datetime, timezone

@pytest.mark.asyncio
async def test_create_an_organization():
    # Create an example organization instance
    org = Organization(
        id="62d65f2506af3feef8fec051",
        name="ALE",
        createdAt=datetime.now(timezone.utc),
        updatedAt=datetime.now(timezone.utc),
        is2FARequired=False,
        imageUrl="",
        countryCode="FR",
        timezone="Europe/Tirane",
        auditHour=130,
        idleTimeout=3600,
        msp="60c8bce3eb5b4155f8b82214",
        upamAuthRecords=30,
        events=30,
        alerts=30,
        wifiRtls=30,
        networkAnalytics=30,
        clientSessions=30,
        clientAnalytics=30,
        auditLogs=7,
        loginAttemps=7,
        iotData=7,
        backupPerDevice=30,
        collectInfo=7,
        configurationBackup=5,
        qoe=30,
        enforceStrongPassword=True,
        enforceStrongPasswordNotifyType="SHOW_MESSAGE_AFTER_LOGIN"
    )

    # Mocked API response (as a dict)
    fake_response = {
        "status": 200,
        "message": "The organization has been successfully fetched.",
        "data": org.model_dump()
    }

    # Create the client and mock `post`
    client = OVCirrusApiClient(base_url="http://mock.api", auth=AsyncMock())
    client.post = AsyncMock(return_value=fake_response)

    # Call the method
    response = await client.createAnOrganization(org)

    # Assertions
    assert response is not None
    assert isinstance(response, ApiResponse)
    assert response.status == 200
    assert response.data.name == "ALE"
    assert response.data.enforceStrongPassword is True



# === api_client/async_client.py ===

import logging
from typing import Optional, Dict, List, Any

import httpx
import backoff
from .auth import Authenticator
from models.generic import ApiResponse
from models.user import UserProfile
from models.organization import Organization
from utilities.model_validator import safe_model_validate

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OVCirrusApiClient:
    def __init__(self, base_url: str, auth: Authenticator):
        self.base_url = base_url.rstrip("/")
        self.auth = auth
        self.client = httpx.AsyncClient()

    def _get_headers(self) -> Dict[str, str]:
        token = self.auth.get_token()
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json"
        }

    @backoff.on_exception(
        backoff.expo,
        (httpx.ConnectError, httpx.ReadTimeout, httpx.RemoteProtocolError),
        max_tries=2
    )

    async def _request(self, method: str, endpoint: str, **kwargs) -> Optional[Any]:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        headers = self._get_headers()
        kwargs.setdefault("headers", headers)

        try:
            response = await self.client.request(method, url, **kwargs)

            # Handle 401 Unauthorized and attempt re-authentication
            if response.status_code == 401:
                logger.warning("Received 401. Attempting re-authentication...")
                if self.auth.force_relogin():
                    headers = self._get_headers()
                    kwargs["headers"] = headers
                    response = await self.client.request(method, url, **kwargs)
                else:
                    logger.error("Re-authentication failed. Exiting.")
                    return None  # Return None after max retries or failure

            response.raise_for_status()

            if response.content:
                try:
                    return response.json()
                except Exception as e:
                    logger.warning(f"Failed to decode JSON from response: {e}")
                    return None
            return None

        except httpx.HTTPStatusError as e:
            logger.warning(f"HTTP error {e.response.status_code}: {e.response.text}")
            return e.response.json()  # Return error details to be handled by the caller
        except Exception as e:
            logger.exception(f"Unhandled exception during API request: {e}")
            raise  # Reraise exception if it's unexpected



    async def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return await self._request("GET", endpoint, params=params)

    async def post(self, endpoint: str, json: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return await self._request("POST", endpoint, json=json)

    async def put(self, endpoint: str, json: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return await self._request("PUT", endpoint, json=json)

    async def delete(self, endpoint: str) -> Optional[Any]:
        return await self._request("DELETE", endpoint)

    async def getUserProfile(self) -> Optional[Any]:
        endpoint = "api/ov/v1/user/profile"
        rawResponse = await self.get(endpoint)
        return safe_model_validate(ApiResponse[UserProfile], rawResponse)

    async def updateUserProfile(self, userProfile: UserProfile) -> Optional[Any]:
        endpoint = "api/ov/v1/user/profile"
        # Convert model to dict instead of JSON string for use with httpx
        rawResponse = await self.put(endpoint, userProfile.model_dump(mode="json"))
        if rawResponse:
            return ApiResponse[UserProfile].model_validate(rawResponse)
        return rawResponse

    async def createAnOrganization(self, organization: Organization) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations"
        rawResponse = await self.post(endpoint, organization)
        if rawResponse:
            return ApiResponse[Organization].model_validate(rawResponse)
        return rawResponse          

    async def getAllUserOrganizations(self) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations"
        rawResponse = await self.get(endpoint)
        if rawResponse:
            return ApiResponse[List[Organization]   ].model_validate(rawResponse)
        return rawResponse           

    async def getOrganizationBasicSettings(self, orgId: str) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations/" + orgId + "/settings/basic"
        rawResponse = await self.get(endpoint)
        if rawResponse:
            return ApiResponse[Organization].model_validate(rawResponse)
        return rawResponse          

    async def getOrganization(self, orgId: str) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations/" + orgId
        rawResponse = await self.get(endpoint)
        if rawResponse:
            return ApiResponse[Organization].model_validate(rawResponse)
        return rawResponse        

    async def updateOrganization(self, orgId: str, organization: Organization) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations/" + orgId
        rawResponse = await self.put(endpoint, organization.model_dump(mode="json"))
        if rawResponse:
            return ApiResponse[Organization].model_validate(rawResponse)
        return rawResponse      

    async def deleteOrganization(self, orgId: str) -> Optional[Any]:
        endpoint = "api/ov/v1/organizations/" + orgId
        rawResponse = await self.delete(endpoint)
        if rawResponse:
            try:
                return ApiResponse[Any].model_validate(rawResponse)
            except:
                return None
        return rawResponse                  
        
    async def close(self):
        await self.client.aclose()



import pytest
import httpx
from httpx import Request, Response
from api_client.async_client import OVCirrusApiClient
from models.user import UserProfile
from models.generic import ApiResponse
from api_client.auth import Authenticator

class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"
    def force_relogin(self) -> bool:
        return True

@pytest.mark.asyncio
async def test_update_user_profile_success():
    user_profile = UserProfile(firstname="John Doe", email="john@example.com")

    mock_response_data = {
        "status": 200,
        "message": "Success",
        "data": user_profile.model_dump()
    }

    async def mock_send(request: Request) -> Response:
        return Response(status_code=200, json=mock_response_data)

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(base_url="https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.updateUserProfile(user_profile)

    assert result is not None
    assert result.status == 200
    assert result.data.firstname == "John Doe"

    await client.close()

@pytest.mark.asyncio
async def test_update_user_profile_invalid_data():
    user_profile = UserProfile(firstname="Invalid", email="bad@example.com")

    mock_response_data = {
        "status": 400,
        "message": "Invalid data",
        "data": None
    }

    async def mock_send(request: Request) -> Response:
        return Response(status_code=400, json=mock_response_data)

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(base_url="https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.updateUserProfile(user_profile)

    assert result.status == 400
    assert result.message == "Invalid data"

    await client.close()

@pytest.mark.asyncio
async def test_update_user_profile_no_response():
    user_profile = UserProfile(firstname="Nobody", email="null@example.com")

    async def mock_send(request: Request) -> Response:
        return Response(status_code=204)

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(base_url="https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.updateUserProfile(user_profile)

    assert result is None

    await client.close()

@pytest.mark.asyncio
async def test_update_user_profile_retry_on_401():
    user_profile = UserProfile(firstname="Retry Guy", email="retry@example.com")

    responses = [401, 200]
    count = 0

    async def mock_send(request: Request) -> Response:
        nonlocal count
        status = responses[count]
        count += 1
        if status == 401:
            return Response(status_code=401, json={"message": "Unauthorized"})
        return Response(status_code=200, json={
            "status": 200,
            "message": "Success",
            "data": user_profile.model_dump()
        })

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(base_url="https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.updateUserProfile(user_profile)

    assert result.status == 200
    assert result.data.firstname == "Retry Guy"

    await client.close()

@pytest.mark.asyncio
async def test_update_user_profile_malformed_response(caplog):
    user_profile = UserProfile(firstname="Mal", email="malformed@example.com")

    async def mock_send(request: Request) -> Response:
        return Response(status_code=200, content=b"{not: valid json")

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(base_url="https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.updateUserProfile(user_profile)

    assert result is None
    assert "Failed to decode JSON" in caplog.text

    await client.close()



from pydantic import BaseModel, Field, Extra,ConfigDict
from typing import Optional
from datetime import datetime

class UserProfile(BaseModel):
    failedTry: Optional[int] = 0
    lockedUntilDate: Optional[int] = 0
    lastLoginDate: Optional[datetime] = None
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    id: Optional[str] = None
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    email: Optional[str] = None
    verified: Optional[bool] = None
    preferredLanguage: Optional[str] = None
    country: Optional[str] = None
    closestRegion: Optional[str] = None
    companyName: Optional[str] = None
    avatarLocation: Optional[str] = None
    address: Optional[str] = None
    city: Optional[str] = None
    zipCode: Optional[str] = None
    phoneNumber: Optional[str] = None
    isTwoFAEnabled: Optional[bool] = None
    faMethod: Optional[str] = None
    tempSecret: Optional[str] = None
    secret: Optional[str] = None
    enforcementPolicy: Optional[str] = None
    accessLevelRole: Optional[str] = None
    # msp: Optional[str] = None

    model_config = ConfigDict(extra="forbid")




from pydantic import BaseModel, Field
from typing import Optional
from datetime import datetime

from typing import Optional, List
from pydantic import BaseModel, Field
from datetime import datetime

class RfProfileDefault(BaseModel):
    id: Optional[int]
    name: Optional[str]
    airtimeFairness2G: Optional[bool] = False
    airtimeFairness5G: Optional[bool] = False
    assignments: Optional[str] = None
    associationRSSIThreshold2: Optional[int] = 0
    associationRSSIThreshold5A: Optional[int] = 0
    associationRSSIThreshold5H: Optional[int] = 0
    associationRSSIThreshold5L: Optional[int] = 0
    associationRSSIThreshold6: Optional[int] = 0
    backgroundScanning: Optional[bool] = True
    band2: Optional[bool] = True
    band5A: Optional[bool] = True
    band5H: Optional[bool] = True
    band5L: Optional[bool] = True
    band6: Optional[bool] = True
    bandSteering: Optional[bool] = False
    bandSteeringForce5g: Optional[bool] = False
    bandSteeringForce6g: Optional[bool] = False
    beaconInterval2: Optional[int] = 100
    beaconInterval5A: Optional[int] = 100
    beaconInterval5H: Optional[int] = 100
    beaconInterval5L: Optional[int] = 100
    beaconInterval6: Optional[int] = 100
    channelDRM5A: Optional[bool] = False
    channelDRM5H: Optional[bool] = False
    channelDRM5L: Optional[bool] = False
    channelDRM6: Optional[bool] = False
    channelList5A: Optional[str] = ""
    channelList5H: Optional[str] = ""
    channelList5L: Optional[str] = ""
    channelList6: Optional[str] = ""
    channelSetting2: Optional[str] = "AUTO"
    channelSetting5A: Optional[str] = "AUTO"
    channelSetting5H: Optional[str] = "AUTO"
    channelSetting5L: Optional[str] = "AUTO"
    channelSetting6: Optional[str] = "AUTO"
    channelWidth2: Optional[str] = "0"
    channelWidth5A: Optional[str] = "0"
    channelWidth5H: Optional[str] = "0"
    channelWidth5L: Optional[str] = "0"
    channelWidth6: Optional[str] = "0"
    clientAwareness2: Optional[bool] = True
    clientAwareness5A: Optional[bool] = True
    clientAwareness5H: Optional[bool] = True
    clientAwareness5L: Optional[bool] = True
    clientAwareness6: Optional[bool] = True
    countryCode: Optional[str] = "AU"
    csa2: Optional[bool] = True
    csa5A: Optional[bool] = True
    csa5H: Optional[bool] = True
    csa5L: Optional[bool] = True
    csa6: Optional[bool] = True
    csaCount2: Optional[int] = 4
    csaCount5A: Optional[int] = 4
    csaCount5H: Optional[int] = 4
    csaCount5L: Optional[int] = 4
    csaCount6: Optional[int] = 4
    defaultSetting: Optional[bool] = False
    description: Optional[str] = ""
    drmInterval: Optional[int] = 6
    drmStartTime: Optional[str] = "00:00"
    drmTimeControl: Optional[bool] = False
    dynamicLoadBalance: Optional[bool] = False
    excludeMacOUILists: Optional[str] = ""
    externalAntennasGain2: Optional[int] = 0
    externalAntennasGain5A: Optional[int] = 0
    externalAntennasGain5H: Optional[int] = 0
    externalAntennasGain5L: Optional[int] = 0
    externalAntennasGain6: Optional[int] = 0
    highEfficiency2: Optional[bool] = True
    highEfficiency5A: Optional[bool] = True
    highEfficiency5H: Optional[bool] = True
    highEfficiency5L: Optional[bool] = True
    highEfficiency6: Optional[bool] = True
    isAPLinked: Optional[bool] = False
    maxTXPower2: Optional[str] = ""
    maxTXPower5A: Optional[str] = ""
    maxTXPower5H: Optional[str] = ""
    maxTXPower5L: Optional[str] = ""
    maxTXPower6: Optional[str] = ""
    minTXPower2: Optional[str] = ""
    minTXPower5A: Optional[str] = ""
    minTXPower5H: Optional[str] = ""
    minTXPower5L: Optional[str] = ""
    minTXPower6: Optional[str] = ""
    muMimo2: Optional[bool] = True
    muMimo5A: Optional[bool] = True
    muMimo5H: Optional[bool] = True
    muMimo5L: Optional[bool] = True
    muMimo6: Optional[bool] = True
    powerSetting2: Optional[str] = "AUTO"
    powerSetting5A: Optional[str] = "AUTO"
    powerSetting5H: Optional[str] = "AUTO"
    powerSetting5L: Optional[str] = "AUTO"
    powerSetting6: Optional[str] = "AUTO"
    profileType: Optional[str] = "NORMAL"
    radioSetting: Optional[List[str]] = []
    roamingRSSIThreshold2: Optional[int] = 0
    roamingRSSIThreshold5A: Optional[int] = 0
    roamingRSSIThreshold5H: Optional[int] = 0
    roamingRSSIThreshold5L: Optional[int] = 0
    roamingRSSIThreshold6: Optional[int] = 0
    scanningChannel: Optional[str] = "workChannel"
    scanningDuration: Optional[int] = 50
    scanningInterval: Optional[int] = 20
    shortGuardInterval2: Optional[bool] = True
    shortGuardInterval5A: Optional[bool] = True
    shortGuardInterval5H: Optional[bool] = True
    shortGuardInterval5L: Optional[bool] = True
    shortGuardInterval6: Optional[bool] = True
    voiceAndVideoAwareness: Optional[bool] = True

class ProvisioningConfigDefault(BaseModel):
    id: Optional[int]
    name: Optional[str]
    siteId: Optional[str]
    advancedAnalyticsIotLocationServerProfileName: Optional[str] = "Default Advanced Analytics"
    advancedAnalyticsSwitch: Optional[bool] = True
    advertisingProtocolMode: Optional[str] = "iBeacon"
    advertisingSwitch: Optional[bool] = False
    apuiPassword: Optional[str] = "*"
    apuiSwitch: Optional[bool] = False
    bleLbsIotLocationServerProfileName: Optional[str] = None
    calledStationId: Optional[str] = None
    cirrusWifiRTLSSwitch: Optional[bool] = False
    dayLightSaveTime: Optional[bool] = False
    description: Optional[str] = None
    eddyStoneURL: Optional[str] = ""
    emissionFrequency: Optional[int] = 200
    eventConfigurationName: Optional[str] = None
    externalPortalCertificate: Optional[str] = None
    hostName: Optional[str] = None
    igmpSnooping: Optional[bool] = False
    iotRadioMode: Optional[str] = "Disable"
    ipv6L3ForwardingSwitch: Optional[bool] = False
    isAPUIPasswordChanged: Optional[bool] = False
    isClientBeTrackingPasswordChanged: Optional[bool] = False
    isSSHRootPasswordChanged: Optional[bool] = False
    isSSHSupportPasswordChanged: Optional[bool] = False
    isSnmpPasswordChanged: Optional[bool] = False
    isTrapPasswordChanged: Optional[bool] = False
    localLdapCertificate: Optional[str] = None
    localRadSecCertificate: Optional[str] = "RadSecCert"
    systemLogDataList: Optional[List[dict]] = [
        {"sysLogId": 226366, "syslogServerIp": "143.209.0.2", "sysLogServerPort": 514, "sysLogTlsSwitch": True, "sysLogCertification": "SyslogCert"}
    ]
    logLevelAPDebug: Optional[str] = "LOG_NOTICE"
    logLevelNetwork: Optional[str] = "LOG_NOTICE"
    logLevelSecurity: Optional[str] = "LOG_NOTICE"
    logLevelSystem: Optional[str] = "LOG_NOTICE"
    logLevelUser: Optional[str] = "LOG_NOTICE"
    logLevelWireless: Optional[str] = "LOG_NOTICE"
    logRemoteSwitch: Optional[bool] = True
    logUploadPeriod: Optional[int] = 1
    nameSpace: Optional[str] = ""
    ntpServerList: Optional[List[str]] = ["pool.ntp.org"]
    ouiAllowList: Optional[List[str]] = []
    password: Optional[str] = "*"
    pmdSwitch: Optional[bool] = False
    profileType: Optional[str] = "DEFAULT"
    redirectProxyServerPort: Optional[int] = 1024
    redirectServerHost: Optional[str] = "dev.portal.ovcirrus.com"
    redirectServerIP: Optional[str] = "172.16.90.109"
    redirectServerIPv6: Optional[str] = ""
    remotePath: Optional[str] = None
    rfProfileName: Optional[str] = "Default RF Profile"
    scanningInterval: Optional[int] = 100
    scanningSwitch: Optional[bool] = False
    serverIP: Optional[str] = None
    serverPort: Optional[int] = 0
    serverType: Optional[str] = "SFTP"
    snmpCommunity: Optional[str] = None
    snmpPassword: Optional[str] = "*"
    snmpStatus: Optional[bool] = False
    snmpUserName: Optional[str] = None
    snmpVersion: Optional[str] = None
    sshRootPassword: Optional[str] = "*"
    sshSupportPassword: Optional[str] = "*"
    sshSwitch: Optional[bool] = False
    syslogServerIp: Optional[str] = None
    syslogServerPort: Optional[int] = 0
    syslogSwitch: Optional[bool] = False
    tftpServer: Optional[str] = None
    timeZone: Optional[str] = "Asia/Singapore (GMT+08:00)"
    trapCommunity: Optional[str] = None
    trapPassword: Optional[str] = "*"
    trapServer: Optional[str] = None
    trapStatus: Optional[bool] = False
    trapUserName: Optional[str] = None
    trapVersion: Optional[str] = None
    txChannel: Optional[List[int]] = [1, 2, 4]
    txPower: Optional[int] = 4
    uploadToServerSwitch: Optional[bool] = False
    userName: Optional[str] = None
    virtualIPAddress: Optional[str] = "192.0.2.0"
    vlanId: Optional[str] = None
    vpnServers: Optional[List[str]] = []
    webServiceCertificate: Optional[str] = None
    wifiIotLocationServerProfileName: Optional[str] = None
    supplicant8021x: Optional[bool] = True
    certificate8021x: Optional[str] = "Cert8021X"

class Organization(BaseModel):
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    id: Optional[str] = None
    name: Optional[str] = ""
    is2FARequired: Optional[bool] = False
    imageUrl: Optional[str] = ""
    countryCode: Optional[str] = ""
    timezone: Optional[str] = ""
    auditHour: Optional[int] = 0
    idleTimeout: Optional[int] = 0
    rfProfileDefault: Optional[RfProfileDefault] = None
    provisioningConfigDefault: Optional[ProvisioningConfigDefault] = None
    msp: Optional[str] = None
    upamAuthRecords: Optional[int] = 0
    events: Optional[int] = 0
    alerts: Optional[int] = 0
    wifiRtls: Optional[int] = 0
    networkAnalytics: Optional[int] = 0
    clientSessions: Optional[int] = 0
    clientAnalytics: Optional[int] = 0
    auditLogs: Optional[int] = 0
    loginAttemps: Optional[int] = 0
    iotData: Optional[int] = 0
    backupPerDevice: Optional[int] = 0
    collectInfo: Optional[int] = 0
    configurationBackup: Optional[int] = 0
    qoe: Optional[int] = 0
    enforceStrongPassword: Optional[bool] = False
    enforceStrongPasswordNotifyType: Optional[str] = "SHOW_MESSAGE_AFTER_LOGIN"



# === api_client/client.py ===

import logging
import requests
from typing import Optional, Any, Dict
from .auth import Authenticator
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class APIClient:
    def __init__(self, base_url: str, auth: Authenticator):
        self.base_url = base_url.rstrip('/')
        self.auth = auth
        self.session = requests.Session()

        # Retry config
        retries = Retry(
            total=3,
            backoff_factor=0.3,
            status_forcelist=[500, 502, 503, 504],
            allowed_methods=["GET", "POST", "PUT", "DELETE", "PATCH"]
        )
        adapter = HTTPAdapter(max_retries=retries)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    def _get_headers(self) -> Dict[str, str]:
        token = self.auth.get_token()
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json"
        }

    def _request(self, method: str, endpoint: str, **kwargs) -> Optional[Any]:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        headers = self._get_headers()
        kwargs.setdefault("headers", headers)

        try:
            response = self.session.request(method, url, **kwargs)

            # If token expired, re-authenticate and retry once
            if response.status_code == 401:
                logger.warning("Received 401. Attempting re-authentication...")
                if self.auth.force_relogin():
                    headers = self._get_headers()
                    kwargs["headers"] = headers
                    response = self.session.request(method, url, **kwargs)
                else:
                    logger.error("Re-authentication failed.")
                    return None

            response.raise_for_status()
            if response.content:
                return response.json()
            return None

        except requests.RequestException as e:
            logger.exception(f"Request to {url} failed.")
            return None

    def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return self._request("GET", endpoint, params=params)

    def post(self, endpoint: str, json: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return self._request("POST", endpoint, json=json)

    def put(self, endpoint: str, json: Optional[Dict[str, Any]] = None) -> Optional[Any]:
        return self._request("PUT", endpoint, json=json)

    def delete(self, endpoint: str) -> Optional[Any]:
        return self._request("DELETE", endpoint)



import pytest
import httpx
from httpx import Response, Request
from typing import Optional
from api_client.async_client import OVCirrusApiClient
from api_client.auth import Authenticator
from models.organization import Organization
from models.generic import ApiResponse
from pydantic import ValidationError

# Dummy authenticator for tests
class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"
    def force_relogin(self) -> bool:
        return False

# Sample payload for a valid organization
valid_org = Organization(
    id="123",
    name="ALE",
    enforceStrongPassword=True,
    enforceStrongPasswordNotifyType="SHOW_MESSAGE_AFTER_LOGIN",
    timezone="Europe/Tirane",
    auditHour=130,
    idleTimeout=3600
)

# Mock successful response
mock_success_response = {
    "status": 200,
    "message": "Updated successfully",
    "data": {
        "id": "123",
        "name": "ALE",
        "enforceStrongPassword": True,
        "enforceStrongPasswordNotifyType": "SHOW_MESSAGE_AFTER_LOGIN",
        "timezone": "Europe/Tirane",
        "auditHour": 130,
        "idleTimeout": 3600
    }
}

# Mock error response
mock_error_response = {
    "status": 400,
    "message": "Invalid data",
    "data": {}
}

# Mock malformed response
mock_malformed_response = {
    "status": 200,
    "message": "Missing data key",
    'data': {
        "sss" : "sss"
    }
}

@pytest.mark.asyncio
async def test_update_organization_success():
    async def mock_send(request: Request) -> Response:
        return Response(200, json=mock_success_response)

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient("https://api.example.com", DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    response: Optional[ApiResponse[Organization]] = await client.updateOrganization("123", valid_org)

    assert response is not None
    assert response.status == 200
    assert response.data.id == "123"
    assert response.data.name == "ALE"

    await client.close()

@pytest.mark.asyncio
async def test_update_organization_invalid_data():
    async def mock_send(request: Request) -> Response:
        return Response(400, json=mock_error_response)

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient("https://api.example.com", DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    response = await client.updateOrganization("123", valid_org)

    assert response is not None
    assert response.status == 400
    assert response.message == "Invalid data"
    assert isinstance(response.data, Organization)

    await client.close()

@pytest.mark.asyncio
async def test_update_organization_no_response():
    async def mock_send(request: Request) -> Response:
        return Response(204)  # No content

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient("https://api.example.com", DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=transport)

    response = await client.updateOrganization("123", valid_org)

    assert response is None
    await client.close()

# @pytest.mark.asyncio
# async def test_update_organization_malformed_response():
#     async def mock_send(request: Request) -> Response:
#         return Response(200, json=mock_malformed_response)

#     transport = httpx.MockTransport(mock_send)
#     client = OVCirrusApiClient("https://api.example.com", DummyAuthenticator())
#     client.client = httpx.AsyncClient(transport=transport)

#     with pytest.raises(ValidationError):  # You can catch a more specific one depending on your validation
#         await client.updateOrganization("123", valid_org)

#     await client.close()



from pydantic import BaseModel
from typing import Generic, TypeVar

class AuthRequest(BaseModel):
    email: str
    password: str
    app_id: str
    app_secret: str

class AuthResponse(BaseModel):
    token_type: str
    expires_in: int  # seconds
    access_token: str





# === api_client/auth.py ===

import requests
import logging
from typing import Optional
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Authenticator:
    def __init__(self):
        self.auth_url: Optional[str] = None
        self.email: Optional[str] = None
        self.password: Optional[str] = None
        self.app_id: Optional[str] = None
        self.app_secret: Optional[str] = None
        self.token: Optional[str] = None
        self.expires_in: Optional[int] = None
        self.token_expiry: Optional[datetime] = None
        self.isAuthenticated: bool = False

    def setURL(self, auth_url: str):
        self.auth_url = auth_url
        return self

    def setCredentials(self, email: str, password: str):
        self.email = email
        self.password = password
        return self  

    def setApp(self, app_id: str, app_secret: str):
        self.app_id = app_id
        self.app_secret = app_secret
        return self

    def build(self):
        if not all([self.auth_url, self.email, self.password, self.app_id, self.app_secret]):
            raise ValueError("Missing authentication configuration.")
        success = self.login()
        if not success:
            raise ValueError("Authentication failed.")
        return self

    def login(self) -> bool:
        logger.info("Authenticating with API...")

        try:
            response = requests.post(self.auth_url, json={
                "email": self.email,
                "password": self.password,
                "appId": self.app_id,
                "appSecret": self.app_secret
            })

            if response.status_code != 200:
                logger.error(f"Auth failed with status {response.status_code}: {response.reason}")
                logger.error(f"Response body: {response.text}")
                self.isAuthenticated = False
                return False

            data = response.json()
            token = data.get("access_token")
            expires_in = data.get("expires_in", 3600)  # default to 1 hour

            if not token:
                logger.error("No access_token returned in authentication response.")
                self.isAuthenticated = False
                return False

            self.token = token
            self.expires_in = expires_in
            self.token_expiry = datetime.utcnow() + timedelta(seconds=expires_in)
            self.isAuthenticated = True

            logger.info("Token acquired successfully. Expires in %s seconds.", expires_in)
            return True

        except requests.RequestException as e:
            logger.exception("Exception during authentication:")
            self.isAuthenticated = False
            return False

    def is_token_expired(self) -> bool:
        if not self.token_expiry:
            return True
        # Refresh 5 minutes before expiry
        return datetime.utcnow() >= self.token_expiry - timedelta(minutes=5)

    def get_token(self) -> Optional[str]:
        if not self.token or self.is_token_expired():
            logger.info("Token missing or expired. Re-authenticating...")
            self.login()
        return self.token

    def force_relogin(self) -> bool:
        logger.info("Forcing re-login...")
        return self.login()



import pytest
import httpx
from httpx import Response, Request
from api_client.async_client import OVCirrusApiClient
from models.organization import Organization
from models.generic import ApiResponse
from api_client.auth import Authenticator

from typing import Optional
from datetime import datetime

# Dummy Authenticator that always returns a token
class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"

    def force_relogin(self) -> bool:
        return True

# Mock response data for multiple organizations
mock_org_data = {
    "status": 200,
    "message": "Success",
    "data": [
        {
            "id": "123",
            "name": "ALE",
            "createdAt": "2022-08-04T12:10:38.058Z",
            "updatedAt": "2022-09-08T15:56:53.407Z",
            "enforceStrongPassword": True,
            "enforceStrongPasswordNotifyType": "SHOW_MESSAGE_AFTER_LOGIN"
        },
        {
            "id": "456",
            "name": "Another Org",
            "createdAt": "2022-01-01T10:00:00.000Z",
            "updatedAt": "2022-02-01T12:00:00.000Z",
            "enforceStrongPassword": False,
            "enforceStrongPasswordNotifyType": "NONE"
        }
    ]
}

@pytest.mark.asyncio
async def test_get_all_user_organizations():
    # Use MockTransport to intercept HTTPX requests
    async def mock_send(request: Request) -> Response:
        return Response(
            status_code=200,
            json=mock_org_data
        )

    transport = httpx.MockTransport(mock_send)

    # Inject httpx.AsyncClient with mock transport
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=transport)

    # Run the method
    response: Optional[ApiResponse[list[Organization]]] = await client.getAllUserOrganizations()

    # Assertions
    assert response is not None
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 2

    first_org = response.data[0]
    assert first_org.name == "ALE"
    assert first_org.enforceStrongPassword is True

    second_org = response.data[1]
    assert second_org.name == "Another Org"
    assert second_org.enforceStrongPassword is False

    await client.close()



import pytest
from unittest.mock import AsyncMock
import httpx
from httpx import Response, Request
from api_client.async_client import OVCirrusApiClient
from models.organization import Organization
from models.generic import ApiResponse
from api_client.auth import Authenticator

# Dummy Authenticator that always returns a token
class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"

    def force_relogin(self) -> bool:
        return True

@pytest.mark.asyncio
async def test_get_organization_valid_response():
    # Mocking a valid response from the API
    raw_response = {
        "status": 200,
        "message": "Success",
        "data": {
            "id": "org123",
            "name": "Test Organization",
            "is2FARequired": True,
            "imageUrl": "http://example.com/image.jpg",
            "countryCode": "US",
            "timezone": "GMT",
            "auditHour": 12,
            "idleTimeout": 60
        }
    }

    # Create an instance of OVCirrusApiClient (replace with actual class)
    obj = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    
    # Mocking the `get` method to return the raw_response
    obj.get = AsyncMock(return_value=raw_response)

    # Call the method
    result = await obj.getOrganization("org123")

    # Verify the result is an ApiResponse and contains the expected data
    assert isinstance(result, ApiResponse)
    assert isinstance(result.data, Organization)
    assert result.data.id == "org123"
    assert result.data.name == "Test Organization"
    assert result.data.is2FARequired is True
    assert result.data.imageUrl == "http://example.com/image.jpg"

@pytest.mark.asyncio
async def test_get_organization_no_response():
    # Mocking the `get` method to return None
    obj = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    
    obj.get = AsyncMock(return_value=None)

    # Call the method
    result = await obj.getOrganization("org123")

    # Assert that the result is None because there's no response
    assert result is None

@pytest.mark.asyncio
async def test_get_organization_invalid_status():
    # Mocking an invalid API response (non-200 status code)
    raw_response = {
        "status": 400,
        "message": "Bad Request",
        "data": {}
    }

    # Create an instance of OVCirrusApiClient
    obj = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    
    # Mock the `get` method to return the raw_response
    obj.get = AsyncMock(return_value=raw_response)

    # Call the method
    result = await obj.getOrganization("org123")

    # Assert that the result is None because the status is not 200
    assert result.status is 400



import pytest
import httpx
from httpx import Request, Response

from api_client.async_client import OVCirrusApiClient
from api_client.auth import Authenticator
from models.generic import ApiResponse
from models.user import UserProfile
from unittest.mock import patch
from typing import Optional
from utilities.model_validator import safe_model_validate

class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"
    def force_relogin(self) -> bool:
        return True


# Mock success data
mock_user_profile_data = {
    "status": 200,
    "message": "OK",
    "data": {
        "id": "u-001",
        "firstname": "John Doe",
        "email": "john@example.com"
    }
}

@pytest.mark.asyncio
async def test_get_user_profile_success():
    async def mock_send(request: Request) -> Response:
        return Response(
            status_code=200,
            json=mock_user_profile_data
        )

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=transport)

    response = await client.getUserProfile()

    assert response is not None
    assert response.status == 200
    assert response.data.firstname == "John Doe"

    await client.close()

@pytest.mark.asyncio
async def test_get_user_profile_invalid_structure():
    async def mock_send(request: Request) -> Response:
        # This data should be missing required fields or have unexpected structure
        return Response(
            status_code=200,
            json={"status": 200, "message": "OK", "data": {"unexpected_field": "oops"}}
        )

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=transport)

    response = await client.getUserProfile()

    # Should fail validation and return None
    assert response is None

    await client.close()




@pytest.mark.asyncio
async def test_get_user_profile_no_response():
    async def mock_send(request: Request) -> Response:
        return Response(status_code=204)

    client = OVCirrusApiClient("https://api.example.com", auth=DummyAuthenticator())
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    result = await client.getUserProfile()
    assert result is None
    await client.close()


@pytest.mark.asyncio
async def test_get_user_profile_unauthorized():
    call_count = 0

    async def mock_send(request: Request) -> Response:
        nonlocal call_count
        call_count += 1
        return Response(
            status_code=401,
            json={"status": 401, "message": "Unauthorized"}
        )

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.getUserProfile()

    assert result is not None  # Because force_relogin() fails after retry
    assert call_count == 2  # First attempt + retry

    await client.close()

@pytest.mark.asyncio
async def test_get_user_profile_retry_success():
    call_count = 0

    async def mock_send(request: Request) -> Response:
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            return Response(
                status_code=401,
                json={"status": 401, "message": "Unauthorized"}
            )
        return Response(
            status_code=200,
            json={
                "status": 200,
                "message": "OK",
                "data": {
                    "id": "u123",
                    "firstname": "Jane Doe",
                    "email": "jane@example.com",
                    "failedTry": 0
                }
            }
        )

    transport = httpx.MockTransport(mock_send)
    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=transport)

    result = await client.getUserProfile()

    assert result is not None
    assert result.status == 200
    assert result.data.firstname == "Jane Doe"
    assert call_count == 2  # Retry happened

    await client.close()


pytest.mark.asyncio
async def test_get_user_profile_malformed_json():
    # Simulate a malformed JSON that was parsed incorrectly into a string (not a dict)
    malformed_data = "{'status': 200, 'message': 'OK', 'data': {'name': 'Jane Doe'}}"  # Not a dict

    with patch("utilities.model_validator.logger") as mock_logger:
        result = safe_model_validate(ApiResponse[UserProfile], malformed_data)

        # safe_model_validate should return None on failure
        assert result is None

        # Ensure a warning was logged
        assert mock_logger.warning.called
        args, _ = mock_logger.warning.call_args
        assert "Validation failed for ApiResponse[UserProfile]" in args[0]






import pytest
import httpx
from httpx import Response, Request
from api_client.async_client import OVCirrusApiClient
from models.organization import Organization
from models.generic import ApiResponse
from api_client.auth import Authenticator

# Dummy Authenticator that always returns a token
class DummyAuthenticator(Authenticator):
    def get_token(self) -> str:
        return "dummy-token"

    def force_relogin(self) -> bool:
        return True


@pytest.mark.asyncio
async def test_delete_organization_success():
    # Mock response
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=200,
            json={"status": 200, "message": "Organization deleted successfully", "data": {"orgId": "123"}}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Assertions for a successful response
    assert response is not None
    assert response.status == 200
    assert response.message == "Organization deleted successfully"
    assert response.data["orgId"] == org_id

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_invalid_id():
    # Mock response for invalid ID
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=400,
            json={"status": 400, "message": "Invalid organization ID", "data": {}}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "invalid-id"
    response = await client.deleteOrganization(org_id)

    # Assertions for a bad request response
    assert response is not None
    assert response.status == 400
    assert response.message == "Invalid organization ID"

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_not_found():
    # Mock response for not found
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=404,
            json={"status": 404, "message": "Organization not found", "data": {}}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Assertions for not found response
    assert response is not None
    assert response.status == 404
    assert response.message == "Organization not found"

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_unauthorized():
    # Mock response for unauthorized
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=401,
            json={"status": 401, "message": "Unauthorized", "data": {}}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Assertions for unauthorized response
    assert response is not None
    assert response.status == 401
    assert response.message == "Unauthorized"

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_server_error():
    # Mock response for server error
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=500,
            json={"status": 500, "message": "Internal server error", "data": {}}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Assertions for server error response
    assert response is not None
    assert response.status == 500
    assert response.message == "Internal server error"

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_no_response():
    # Mock response for no response body
    async def mock_send(request: Request) -> Response:
        return httpx.Response(
            status_code=200,
            json=None
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Assertions for no response returned
    assert response is None

    await client.close()


@pytest.mark.asyncio
async def test_delete_organization_malformed_response():
    # Mock response for malformed JSON or invalid structure
    async def mock_send(request: Request) -> Response:
        # Returning a response with an unexpected structure
        return httpx.Response(
            status_code=200,
            json={"unexpected_field": "value"}
        )

    client = OVCirrusApiClient(
        base_url="https://api.example.com",
        auth=DummyAuthenticator()
    )
    client.client = httpx.AsyncClient(transport=httpx.MockTransport(mock_send))

    org_id = "123"
    response = await client.deleteOrganization(org_id)

    # Ensure response is properly handled, and assert the unexpected_field is present
    assert response is not None
    

    await client.close()

